/**
 * ==============================================
 *  DQL : Descartes Query Language
 * ==============================================
 *
 * (c) Copyright 2014-2016, by Juergen Walter, Fabian Gorsler, Fabian Brosig and other contributors.
 *
 * Project Info:   http://descartes.tools/dql
 *
 * All rights reserved. This software is made available under the terms of the
 * Eclipse Public License (EPL) v1.0 as published by the Eclipse Foundation
 * http://www.eclipse.org/legal/epl-v10.html
 *
 * This software is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
 * or FITNESS FOR A PARTICULAR PURPOSE. See the Eclipse Public License (EPL)
 * for more details.
 *
 * You should have received a copy of the Eclipse Public License (EPL)
 * along with this software; if not visit http://www.eclipse.org or write to
 * Eclipse Foundation, Inc., 308 SW First Avenue, Suite 110, Portland, 97204 USA
 * Email: license (at) eclipse.org
 *
 * [Java is a trademark or registered trademark of Sun Microsystems, Inc.
 * in the United States and other countries.]
 */
package tools.descartes.dql.connector.nop;

import java.math.BigDecimal;
import java.util.ArrayList;
import java.util.Date;
import java.util.List;
import java.util.Properties;

import org.eclipse.emf.ecore.util.EcoreUtil;

import tools.descartes.dql.connector.PerformanceMetricsQueryConnector;
import tools.descartes.dql.models.mapping.domain.AbstractSeriesResult;
import tools.descartes.dql.models.mapping.domain.ConfidenceInterval;
import tools.descartes.dql.models.mapping.domain.ContinuousSeriesResult;
import tools.descartes.dql.models.mapping.domain.DecimalResult;
import tools.descartes.dql.models.mapping.domain.DiscreteSeriesResult;
import tools.descartes.dql.models.mapping.domain.DoF;
import tools.descartes.dql.models.mapping.domain.DomainFactory;
import tools.descartes.dql.models.mapping.domain.Entity;
import tools.descartes.dql.models.mapping.domain.OnlineResult;
import tools.descartes.dql.models.mapping.domain.Probe;
import tools.descartes.dql.models.mapping.domain.ResultList;
import tools.descartes.dql.models.mapping.domain.SeriesResultElement;
import tools.descartes.dql.models.mapping.domain.StatType;
import tools.descartes.dql.models.mapping.domain.StringResult;
import tools.descartes.dql.models.mapping.mapping.EntityMapping;

public class PerformanceMetricsQueryConnectorImpl extends AbstractNopQueryConnector
		implements PerformanceMetricsQueryConnector {
	ArrayList<String> constraintNames = new ArrayList<String>();
	ArrayList<String> doFConstraintNames = new ArrayList<String>();
	ArrayList<String> explorationStrategyNames = new ArrayList<String>();
	private String constraint = null;
	private String doFConstraint = null;
	private String exploratationStrategy = null;
	private Properties explorationStrategyProperties;
	private final DomainFactory domainFactory = DomainFactory.eINSTANCE;

	public PerformanceMetricsQueryConnectorImpl() {
		super();

		loadConstraintNames();
		loadDoFConstraintNames();
		loadExplorationStrategyNames();
	}

	@Override
	public String getConstraint() {
		return constraint;
	}

	@Override
	public String getDoFConstraint() {
		return doFConstraint;
	}

	@Override
	public String getDoFExplorationStrategy() {
		return exploratationStrategy;
	}

	@Override
	public Properties getDoFExplorationStrategyProperties() {
		return explorationStrategyProperties;
	}

	@Override
	public List<String> getSupportedConstraintNames() {
		return constraintNames;
	}

	@Override
	public List<String> getSupportedDoFConstraintNames() {
		return doFConstraintNames;
	}

	@Override
	public List<String> getSupportedExplorationStrategyNames() {
		return explorationStrategyNames;
	}

	@Override
	public List<String> getSupportedTimeUnits() {
		// TODO Auto-generated method stub
		return null;
	}

	private void loadConstraintNames() {
		constraintNames.add("DEFAULT");
		constraintNames.add("FAST");
		constraintNames.add("ACCURATE");
		constraintNames.add("TIME_1s");
		constraintNames.add("TIME_5s");
		constraintNames.add("ACCURACY_95");
		constraintNames.add("ACCURACY_99");
	}

	private void loadDoFConstraintNames() {
		doFConstraintNames.add("DEFAULT");
		doFConstraintNames.add("FAST");
		doFConstraintNames.add("ACCURATE");
		doFConstraintNames.add("TIME_1s");
		doFConstraintNames.add("TIME_5s");
		doFConstraintNames.add("ACCURACY_95");
		doFConstraintNames.add("ACCURACY_99");
	}

	private void loadExplorationStrategyNames() {
		explorationStrategyNames.add("FULL");
		explorationStrategyNames.add("RANDOM");
	}

	@Override
	public List<EntityMapping> requestDoFMetrics(final EntityMapping request) {
		fakeProgress();
		final ArrayList<EntityMapping> results = new ArrayList<EntityMapping>();

		for (@SuppressWarnings("unused")
		final DoF doF : request.getDoF()) {
			results.add(requestMetrics(request));
		}

		return results;
	}

	@Override
	public EntityMapping requestMetrics(final EntityMapping request) {
		fakeProgressIfNotDoF();
		final EntityMapping response = EcoreUtil.copy(request);
		final ArrayList<Entity> entities = new ArrayList<Entity>();
		entities.addAll(response.getResources());
		entities.addAll(response.getServices());

		for (final Entity entity : entities) {
			for (final Probe probe : entity.getProbes()) {
				for (StatType rt : probe.getStatTypes()) {
					if (probe.getMetricName().startsWith("discreteSeries")) {
						final DiscreteSeriesResult result = domainFactory.createDiscreteSeriesResult();
						result.setValid(true);
						result.setXAxisDescription("x");
						result.setYAxisDescription("y");
						for (int i = 0; i < 9; i++) {
							SeriesResultElement elem = domainFactory.createSeriesResultElement();
							ConfidenceInterval ci = domainFactory.createConfidenceInterval();
							ci.setAccuracy(BigDecimal.valueOf(0.95));
							ci.setIntervalHalfLength(BigDecimal.valueOf(0.2));
							elem.setConfidenceInterval(ci);
							elem.setX(BigDecimal.valueOf(i));
							elem.setY(BigDecimal.valueOf(Math.cos(i) + 3));
							result.getElements().add(elem);
						}
						rt.setResult(result);
					} else if (probe.getMetricName().startsWith("continuousSeries")) {
						rt.setResult(createContinuousSeriesResult(2, probe.getMetricName()));
					} else if (probe.getMetricName().startsWith("string")) {
						final StringResult result = domainFactory.createStringResult();
						result.setValid(true);
						result.setValue("This is a String");
						rt.setResult(result);
					} else if (probe.getMetricName().startsWith("listOfContinuousSeries")) {
						final ResultList<ContinuousSeriesResult> result = domainFactory
								.createResultList(ContinuousSeriesResult.class);
						result.setValid(true);
						result.getResults().add(createContinuousSeriesResult(3, "Series1"));
						result.getResults().add(createContinuousSeriesResult(10, "Series2"));
						result.getResults().add(createContinuousSeriesResult(7, "Series3"));
						rt.setResult(result);
					} else if (probe.getMetricName().startsWith("onlineSeries")) {
						final OnlineResult outerResult = domainFactory.createOnlineResult();
						final AbstractSeriesResult result = domainFactory.createContinuousSeriesResult();
						result.setValid(true);
						result.setXAxisDescription("x");
						result.setYAxisDescription("y");
						outerResult.setResult(result);
						Thread updater = new OnlineSeriesUpdateThread(result, 100, 1000);
						updater.start();
						rt.setResult(outerResult);
					} else {
						final DecimalResult result = domainFactory.createDecimalResult();
						result.setValid(true);

						ConfidenceInterval ci = domainFactory.createConfidenceInterval();
						ci.setAccuracy(BigDecimal.valueOf(0.95));
						ci.setIntervalHalfLength(BigDecimal.valueOf(0.2));
						result.setConfidenceInterval(ci);
						result.setValue(new BigDecimal(1.234f));

						rt.setResult(result);
					}
				}
			}
		}

		return response;
	}

	private ContinuousSeriesResult createContinuousSeriesResult(int j, String name) {
		final ContinuousSeriesResult result = domainFactory.createContinuousSeriesResult();
		result.setValid(true);
		result.setXAxisDescription("x");
		result.setYAxisDescription("y");
		for (int i = 0; i < 9; i++) {
			SeriesResultElement elem = domainFactory.createSeriesResultElement();
			ConfidenceInterval ci = domainFactory.createConfidenceInterval();
			ci.setAccuracy(BigDecimal.valueOf(0.95));
			ci.setIntervalHalfLength(BigDecimal.valueOf(0.2));
			elem.setConfidenceInterval(ci);
			elem.setX(BigDecimal.valueOf(i));
			elem.setY(BigDecimal.valueOf(Math.cos(i) + j));
			result.getElements().add(elem);
		}
		return result;
	}

	@Override
	public void setConstraint(final String constraint) {
		// Although being not relevant for results, the value is stored and
		// returned by the getter, e.g. for regression testing.
		if (constraintNames.contains(constraint)) {
			this.constraint = constraint;
		} else {
			this.constraint = null;
		}
	}

	@Override
	public void setDoFConstraint(final String doFConstraint) {
		if (doFConstraintNames.contains(doFConstraint)) {
			this.doFConstraint = doFConstraint;
		} else {
			this.doFConstraint = null;
		}
	}

	@Override
	public void setDoFExplorationStrategy(final String strategy, final Properties properties) {
		if (explorationStrategyNames.contains(strategy)) {
			exploratationStrategy = strategy;
		} else {
			exploratationStrategy = null;
		}

		explorationStrategyProperties = properties;
	}

	@Override
	public void setObservationPeriod(final Date start, final Date end, final long sampleSize) {
		// TODO Auto-generated method stub

	}

	@Override
	public void setObservationPeriod(final long amount, final String unit, final long sampleSize) {
		// TODO Auto-generated method stub

	}

	@Override
	public void setObservationReference(final String type, final String identifier, final long sampleSize) {

	}

	@Override
	public boolean supportsConstraintRequests() {
		return true;
	}

	@Override
	public boolean supportsDoFRequests() {
		return true;
	}

	@Override
	public boolean supportsMetricRequests() {
		return true;
	}

	@Override
	public boolean supportsObservationPeriods() {
		return false;
	}

	@Override
	public void reset() {
		// Nothing to do here
	}
	
	private void fakeProgressIfNotDoF() {
		StackTraceElement[] stackTrace = Thread.currentThread().getStackTrace();
		boolean isCalledByDoF = false;
		for (int i = 0; i < stackTrace.length; i++) {
			if (stackTrace[i].getMethodName().equals("requestDoFMetrics")) {
				isCalledByDoF = true;
				break;
			}
		}

		if (!isCalledByDoF) {
			fakeProgress();
		}
	}
}
