/**
 * ==============================================
 *  DQL : Descartes Query Language
 * ==============================================
 *
 * (c) Copyright 2014-2016, by Juergen Walter, Fabian Gorsler, Fabian Brosig and other contributors.
 *
 * Project Info:   http://descartes.tools/dql
 *
 * All rights reserved. This software is made available under the terms of the
 * Eclipse Public License (EPL) v1.0 as published by the Eclipse Foundation
 * http://www.eclipse.org/legal/epl-v10.html
 *
 * This software is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
 * or FITNESS FOR A PARTICULAR PURPOSE. See the Eclipse Public License (EPL)
 * for more details.
 *
 * You should have received a copy of the Eclipse Public License (EPL)
 * along with this software; if not visit http://www.eclipse.org or write to
 * Eclipse Foundation, Inc., 308 SW First Avenue, Suite 110, Portland, 97204 USA
 * Email: license (at) eclipse.org
 *
 * [Java is a trademark or registered trademark of Sun Microsystems, Inc.
 * in the United States and other countries.]
 */
package tools.descartes.dql.core.engine.interpreter;

import java.util.ArrayList;
import java.util.List;

import org.apache.log4j.Logger;
import org.eclipse.emf.ecore.EObject;
import org.eclipse.xtext.serializer.impl.Serializer;
import org.osgi.framework.BundleContext;
import org.osgi.framework.ServiceReference;

import com.google.inject.Injector;

import tools.descartes.dql.connector.ConnectorProvider;
import tools.descartes.dql.connector.ModelStructureQueryConnector;
import tools.descartes.dql.connector.PerformanceIssueQueryConnector;
import tools.descartes.dql.connector.PerformanceMetricsQueryConnector;
import tools.descartes.dql.connector.QueryConnector;
import tools.descartes.dql.connector.QueryProgressListener;
import tools.descartes.dql.core.engine.Activator;
import tools.descartes.dql.core.engine.util.DQLLogger;
import tools.descartes.dql.core.registry.ConnectorRegistry;
import tools.descartes.dql.models.mapping.mapping.EntityMapping;

public abstract class AbstractQueryInterpreter<Q, qC> implements QueryInterpreter {
	protected Logger log = DQLLogger.getLogger(this.getClass().getName());
	private BundleContext context = Activator.getContext();
	private qC queryConnector = null;
	private List<EntityMapping> results = null;
	private Q query = null;
	private String modelFamily = null;
	private ServiceReference<ConnectorRegistry> registryServiceRef;
	private final Class<qC> myQueryConnectorClass = getQueryConnectorClass();
	private List<QueryProgressListener> listeners = new ArrayList<QueryProgressListener>();

	public AbstractQueryInterpreter(final Q query) {
		this.query = query;
	}

	public AbstractQueryInterpreter(final Q query, List<QueryProgressListener> listeners) {
		this(query);
		this.listeners = listeners;
	}

	public BundleContext getContext() {
		return context;
	}

	public String getModelFamily() {
		return modelFamily;
	}

	public Q getQuery() {
		return query;
	}

	public qC getQueryConnector() {
		return queryConnector;
	}

	protected abstract Class<qC> getQueryConnectorClass();

	// TODO: Should be implemented with type-safety
	@SuppressWarnings("unchecked")
	private qC getQueryConnectorInstance(final ConnectorProvider connectorProvider) {
		log.debug("Trying to create QueryConnector of type " + myQueryConnectorClass.getName());

		if (myQueryConnectorClass == ModelStructureQueryConnector.class) {
			queryConnector = (qC) connectorProvider.createModelStructureQueryConnectorInstance();
		} else if (myQueryConnectorClass == PerformanceMetricsQueryConnector.class) {
			queryConnector = (qC) connectorProvider.createPerformanceMetricsQueryConnectorInstance();
		} else if (myQueryConnectorClass == PerformanceIssueQueryConnector.class) {
			queryConnector = (qC) connectorProvider.createPerformanceIssueQueryConnectorInstance();
		}

		return queryConnector;
	}

	@Override
	public EntityMapping getResult() {
		if (results == null) {
			return null;
		}

		return results.get(0);
	}

	@Override
	public List<EntityMapping> getResults() {
		return results;
	}

	@Override
	public boolean hasMultipleResults() {
		if (results == null || results.size() < 2) {
			return false;
		}

		return true;
	}

	public boolean initQueryConnector() {
		// No null check, as the registry is a dependency and must be available
		// if this bundle is active
		final ConnectorRegistry connectorRegistry = lookupConnectorRegistryInstance();

		final ServiceReference<ConnectorProvider> connectorProviderRef = lookupConnectorProviderReference(
				connectorRegistry, getModelFamily());

		// Might happen even caused by a typo in the query...
		if (connectorProviderRef == null) {
			log.error("No ConnectorProvider found for family " + getModelFamily() + ", aborting.");
			return false;
		}

		final ConnectorProvider connectorProvider = context.getService(connectorProviderRef);

		// Might happen due to dynamics in the run-time and no direct bundle
		// dependency
		if (connectorProvider == null) {
			log.error("Could not instantiate the ConnectorProvider, aborting.");
			return false;
		}

		queryConnector = getQueryConnectorInstance(connectorProvider);

		// Erroneous implementations of the ConnectorProvider might cause this
		if (queryConnector == null) {
			log.error("Although advertised, the ConnectorProvider can not offer a " + myQueryConnectorClass
					+ ", aborting.");
			return false;
		}

		QueryConnector connector = (QueryConnector) queryConnector;
		for (QueryProgressListener listener : listeners) {
			connector.addQueryProgressListener(listener);
		}

		return true;
	}

	protected ServiceReference<ConnectorProvider> lookupConnectorProviderReference(final ConnectorRegistry registry,
			final String family) {
		return registry.requestServiceReference(family, myQueryConnectorClass);
	}

	private ConnectorRegistry lookupConnectorRegistryInstance() {
		registryServiceRef = context.getServiceReference(ConnectorRegistry.class);
		return context.getService(registryServiceRef);
	}

	public void setContext(final BundleContext context) {
		this.context = context;
	}

	public void setModelFamily(final String modelFamily) {
		this.modelFamily = modelFamily;
	}

	public void setQuery(final Q query) {
		this.query = query;
	}

	public void setResult(final EntityMapping result) {
		if (results == null) {
			results = new ArrayList<EntityMapping>();
		}

		results.add(result);
	}

	@SuppressWarnings("restriction")
	public String getQueryAsString() {
		Injector injector = Activator.getDqlInjector();
		Serializer serializer = injector.getInstance(Serializer.class);
		return serializer.serialize((EObject) query);
	}

	public void setResults(final List<EntityMapping> results) {
		this.results = results;
	}
}
