/**
 * ==============================================
 *  DQL : Descartes Query Language
 * ==============================================
 *
 * (c) Copyright 2014-2016, by Juergen Walter, Fabian Gorsler, Fabian Brosig and other contributors.
 *
 * Project Info:   http://descartes.tools/dql
 *
 * All rights reserved. This software is made available under the terms of the
 * Eclipse Public License (EPL) v1.0 as published by the Eclipse Foundation
 * http://www.eclipse.org/legal/epl-v10.html
 *
 * This software is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
 * or FITNESS FOR A PARTICULAR PURPOSE. See the Eclipse Public License (EPL)
 * for more details.
 *
 * You should have received a copy of the Eclipse Public License (EPL)
 * along with this software; if not visit http://www.eclipse.org or write to
 * Eclipse Foundation, Inc., 308 SW First Avenue, Suite 110, Portland, 97204 USA
 * Email: license (at) eclipse.org
 *
 * [Java is a trademark or registered trademark of Sun Microsystems, Inc.
 * in the United States and other countries.]
 */
package tools.descartes.dql.core.engine.interpreter;

import static tools.descartes.dql.core.engine.observations.ObservationInterpreterFactory.createInterpreter;
import static tools.descartes.dql.core.engine.util.DoFHelper.processVariationClause;
import static tools.descartes.dql.core.engine.util.QueryParserHelper.toFlatConfigurationPropertiesClause;

import java.util.HashMap;
import java.util.LinkedList;
import java.util.List;
import java.util.Map;
import java.util.Properties;

import org.eclipse.emf.common.util.EList;
import org.eclipse.emf.ecore.EObject;

import tools.descartes.dql.connector.PerformanceMetricsQueryConnector;
import tools.descartes.dql.connector.QueryProgressListener;
import tools.descartes.dql.core.engine.aggregates.AggregateController;
import tools.descartes.dql.core.engine.aggregates.AggregateControllerImpl;
import tools.descartes.dql.core.engine.aggregates.AggregateProcessorImpl;
import tools.descartes.dql.core.engine.observations.ObservationInterpreter;
import tools.descartes.dql.core.engine.observations.SampleInterpreter;
import tools.descartes.dql.core.engine.query.QueryExecutionEngine;
import tools.descartes.dql.core.engine.query.QueryExecutionEngineImpl;
import tools.descartes.dql.lang.descartesQL.AggregateMetricClause;
import tools.descartes.dql.lang.descartesQL.AliasClause;
import tools.descartes.dql.lang.descartesQL.ConfigurationPropertiesClause;
import tools.descartes.dql.lang.descartesQL.ConstraintClause;
import tools.descartes.dql.lang.descartesQL.DescartesQL;
import tools.descartes.dql.lang.descartesQL.DescartesQLFactory;
import tools.descartes.dql.lang.descartesQL.DoFClause;
import tools.descartes.dql.lang.descartesQL.DoFReference;
import tools.descartes.dql.lang.descartesQL.EntityReference;
import tools.descartes.dql.lang.descartesQL.ExplorationStrategyClause;
import tools.descartes.dql.lang.descartesQL.ForClause;
import tools.descartes.dql.lang.descartesQL.ListMetricsQuery;
import tools.descartes.dql.lang.descartesQL.ListQuery;
import tools.descartes.dql.lang.descartesQL.MetricClause;
import tools.descartes.dql.lang.descartesQL.MetricClauseType;
import tools.descartes.dql.lang.descartesQL.MetricReference;
import tools.descartes.dql.lang.descartesQL.MetricReferenceClauses;
import tools.descartes.dql.lang.descartesQL.ModelFamily;
import tools.descartes.dql.lang.descartesQL.ModelLocation;
import tools.descartes.dql.lang.descartesQL.ModelReferenceClause;
import tools.descartes.dql.lang.descartesQL.ObserveClause;
import tools.descartes.dql.lang.descartesQL.PerformanceMetricsQuery;
import tools.descartes.dql.lang.descartesQL.SelectQuery;
import tools.descartes.dql.lang.descartesQL.UsingClause;
import tools.descartes.dql.lang.descartesQL.VaryingClause;
import tools.descartes.dql.lang.descartesQL.WithClause;
import tools.descartes.dql.models.mapping.domain.DoF;
import tools.descartes.dql.models.mapping.domain.DomainFactory;
import tools.descartes.dql.models.mapping.domain.Entity;
import tools.descartes.dql.models.mapping.domain.Probe;
import tools.descartes.dql.models.mapping.domain.Resource;
import tools.descartes.dql.models.mapping.domain.Service;
import tools.descartes.dql.models.mapping.domain.StatType;
import tools.descartes.dql.models.mapping.mapping.EntityMapping;
import tools.descartes.dql.models.mapping.mapping.MappingFactory;

/**
 * Interpreter for SELECT queries
 * 
 * @author Fabian Gorsler <fabian.gorsler@student.kit.edu>
 */
public class PerformanceMetricsQueryInterpreter extends
		AbstractQueryInterpreter<PerformanceMetricsQuery, PerformanceMetricsQueryConnector> {

	private final MappingFactory mappingFactory = MappingFactory.eINSTANCE;
	private final DomainFactory domainFactory = DomainFactory.eINSTANCE;
	private final DescartesQLFactory dqlFactory = DescartesQLFactory.eINSTANCE;
	private EntityMapping result;

	public PerformanceMetricsQueryInterpreter(final PerformanceMetricsQuery query) {
		super(query);
	}
	
	public PerformanceMetricsQueryInterpreter(final PerformanceMetricsQuery query, List<QueryProgressListener> listeners) {
		super(query, listeners);
	}

	@Override
	protected Class<PerformanceMetricsQueryConnector> getQueryConnectorClass() {
		return PerformanceMetricsQueryConnector.class;
	}

	@Override
	public boolean interpretQuery() {
		SelectQuery selectQuery = null;

		if (getQuery() instanceof SelectQuery) {
			selectQuery = (SelectQuery) getQuery();
			
			final UsingClause usingClause = selectQuery.getUsingClause();
			final ModelReferenceClause modelReference = usingClause.getModelReference();
			final ModelFamily family = modelReference.getFamily();
			final String familyIdentifier = family.getFamily();
			
			setModelFamily(familyIdentifier);

			if (!initQueryConnector()) {
				String exception = "Could not load Connector for family " + getModelFamily() + ", aborting.";
				log.error(exception);
				throw new IllegalStateException(exception);
			}

			// Reset before work, allowing independent caching mechanisms in
			// the DQL Connector
			getQueryConnector().reset();

			return interpretSelectQuery(selectQuery);
		}

		log.error("Query type not supported");
		return false;
	}

	/**
	 * Entry point for interpreting SELECT
	 * 
	 * @param selectQuery
	 *            The query to be interpreted
	 * @return True, if interpreting the query was successful
	 */
	private boolean interpretSelectQuery(final SelectQuery selectQuery) {
		final PerformanceMetricsQueryConnector queryConnector = getQueryConnector();
		final Map<String, Entity> aliasMap = new HashMap<String, Entity>();
		final EntityMapping request = mappingFactory.createEntityMapping();
		request.setQueryAsString(getQueryAsString());
		final AggregateController aggregates = new AggregateControllerImpl();
		final UsingClause usingClause = selectQuery.getUsingClause();
		final ModelReferenceClause modelReference = usingClause.getModelReference();
		final ModelLocation model = modelReference.getModel();
		final String location = model.getLocation();

		if (!queryConnector.supportsMetricRequests()) {
			log.error("Query connector has no support for Performance Metrics, aborting.");
			return false;
		}

		// process using
		request.setModelLocation(location);

		// process for
		if (!processForClause(selectQuery, aliasMap, request)) {
			return false;
		}

		// process select
		if (!processSelectClause(selectQuery, aliasMap, aggregates)) {
			return false;
		}

		// process constrained query
		if (queryConnector.supportsConstraintRequests()) {
			processConstraintClause(selectQuery, queryConnector);
		}

		// process observe clause
		processObserveClause(selectQuery, queryConnector);

		final boolean hasAggregates = aggregates.hasAggregates();
		final DoFClause dofClause = selectQuery.getDofClause();

		if (dofClause != null) {
			// Request results with the interpretation of DoF

			if (!queryConnector.supportsDoFRequests()) {
				log.error("Query connector has no support for DoF Queries, aborting.");
				return false;
			}

			processDoFClause(queryConnector, selectQuery, request);

			if (hasAggregates) {
				aggregates.preProcess(request, aliasMap);
			}

			setResults(queryConnector.requestDoFMetrics(request));

			if (hasAggregates) {
				for (final EntityMapping response : getResults()) {
					aggregates.postProcess(response);
				}
			}
		} else {
			// Request results without the interpretation of DoF

			if (hasAggregates) {
				aggregates.preProcess(request, aliasMap);
			}
			setResult(queryConnector.requestMetrics(request));

			if (hasAggregates) {
				aggregates.postProcess(getResult());
			}
		}

		return true;
	}

	/**
	 * Process a constraint clause in best-effort manner
	 * 
	 * @param selectQuery
	 *            The query to be interpreted
	 * @param queryConnector
	 *            The query connector to execute the query
	 */
	private void processConstraintClause(final SelectQuery selectQuery,
			final PerformanceMetricsQueryConnector queryConnector) {
		final ConstraintClause constraintClause = selectQuery.getConstraintClause();

		if (constraintClause != null) {
			final String constraint = constraintClause.getType();
			queryConnector.setConstraint(constraint);
		} else {
			queryConnector.setConstraint(null);
		}
	}

	/**
	 * Process a EVALUATE DOF clause with Degrees-of-Freedom
	 * 
	 * @param queryConnector
	 *            The query connector to execute the query
	 * @param selectQuery
	 *            The query to be interpreted
	 * @param request
	 *            The request to be sent to the query connector
	 */
	private void processDoFClause(final PerformanceMetricsQueryConnector queryConnector, final SelectQuery selectQuery,
			final EntityMapping request) {
		final DoFClause doFClause = selectQuery.getDofClause();
		final VaryingClause varyingClause = doFClause.getVaryingClause();
		final ExplorationStrategyClause explorationStrategyClause = doFClause.getExplorationStrategy();
		final ConstraintClause constraintClause = selectQuery.getConstraintClause();
		final WithClause withClause = doFClause.getWithClause();

		if (withClause != null) {
			request.setDoFModelLocation(withClause.getModelReference().getModel().getLocation());
		}

		if (explorationStrategyClause != null) {
			processExplorationStrategyClause(queryConnector, explorationStrategyClause);
		}

		if (constraintClause != null) {
			queryConnector.setDoFConstraint(constraintClause.getType());
		}

		if (varyingClause != null) {
			processVaryingClause(request, varyingClause);
		}

	}

	/**
	 * Process a CONSTRAINED AS clause in best-effort manner
	 * 
	 * @param queryConnector
	 *            The query connector to execute the query
	 * @param explorationStrategyClause
	 */
	private void processExplorationStrategyClause(final PerformanceMetricsQueryConnector queryConnector,
			final ExplorationStrategyClause explorationStrategyClause) {
		final ConfigurationPropertiesClause configurationProperties = explorationStrategyClause.getParameterClause();
		final Properties explorationStrategyProperties = toFlatConfigurationPropertiesClause(configurationProperties);
		queryConnector.setDoFExplorationStrategy(explorationStrategyClause.getName(), explorationStrategyProperties);
	}

	/**
	 * Process a FOR clause
	 * 
	 * @param queryConnector
	 *            The query connector to execute the query
	 * @param selectQuery
	 *            The query to be interpreted
	 * @param aliasMap
	 *            A index structure with mappings from aliases to entities
	 * @param request
	 *            The request to be sent to the query connector
	 * 
	 * @return True, if the clause was consistent and could be interpreted
	 */
	private boolean processForClause(final SelectQuery selectQuery, final Map<String, Entity> aliasMap,
			final EntityMapping request) {
		final List<Resource> resources = request.getResources();
		final List<Service> services = request.getServices();
		final List<EntityReference> entityRefs = selectQuery.getForClause().getEntityReferenceClause()
				.getEntityReferences();

		for (final EntityReference ref : entityRefs) {
			Entity entity = null;

			switch (ref.getType()) {
			case RESOURCE:
				entity = domainFactory.createResource();
				resources.add((Resource) entity);
				break;
			case SERVICE:
				entity = domainFactory.createService();
				services.add((Service) entity);
				break;
			default:
				log.error("Invalid entity type found, aborting.");
				return false;
			}

			final AliasClause aliasClause = ref.getAlias();

			if (aliasClause != null) {
				entity.setAlias(aliasClause.getAlias());
			}

			entity.setIdentifier(ref.getIdentifier());

			if (aliasMap.put(entity.getAlias(), entity) != null) {
				log.error("Duplicate aliases in FOR, aborting.");
				return false;
			}
		}

		return true;
	}

	/**
	 * TODO: Alpha stage, skip in code review
	 * 
	 * @param selectQuery
	 * @param queryConnector
	 */
	private void processObserveClause(final SelectQuery selectQuery,
			final PerformanceMetricsQueryConnector queryConnector) {
		final ObserveClause observeClause = selectQuery.getObserveClause();
		final boolean supportsObservationPeriods = queryConnector.supportsObservationPeriods();

		if (observeClause == null) {
			return;
		}

		if (!supportsObservationPeriods) {
			log.warn("Observation period set, but not supported by the query connector. Ignoring OBSERVE.");
			return;
		}

		final ObservationInterpreter oi = createInterpreter(observeClause.getObservation());
		final SampleInterpreter si = new SampleInterpreter(observeClause.getSampleClause());

		oi.configureConnector(queryConnector, si.getSampleSize());
	}

	/**
	 * Process a SELECT clause an resolve all metrics and aggregates to be
	 * calculated
	 * 
	 * @param selectQuery
	 *            The query to be interpreted
	 * @param aliasMap
	 *            A index structure with mappings from aliases to entities
	 * @param aggregates
	 *            A controller for the processing of aggregates
	 * 
	 * @return True, if the clause was consistent and could be interpreted
	 */
	private boolean processSelectClause(final SelectQuery selectQuery, final Map<String, Entity> aliasMap,
			final AggregateController aggregates) {
		final MetricReferenceClauses metricRefClauses = selectQuery.getMetricReferenceClauses();
		final EList<MetricClauseType> metricClauseTypes = metricRefClauses.getMetricClauseTypes();
		initializeEntityMapping(selectQuery);

		for (final MetricClauseType ref : metricClauseTypes) {
			final Probe probe = domainFactory.createProbe();
			Entity entity = null;

			if (ref instanceof MetricClause) {
				// Found a plain metric
				final MetricClause metricClause = (MetricClause) ref;

				final MetricReference metricRef = metricClause.getMetricReference();
				final String entityRef = metricRef.getEntityReference();
				final String metricRef2 = metricRef.getMetricReference();
				String statTypeRef = null;
				if (metricRef.getStatTypeReference() != null)
					statTypeRef = metricRef.getStatTypeReference();
				else
					statTypeRef = getDefaultStatType(entityRef, metricRef2);

				entity = aliasMap.get(entityRef);
				probe.setMetricName(metricRef2);
				final StatType st = domainFactory.createStatType();
				st.setTypeName(statTypeRef);
				probe.getStatTypes().add(st);
			} else if (ref instanceof AggregateMetricClause) {
				// Found an aggregate
				aggregates.registerAggregate(new AggregateProcessorImpl((AggregateMetricClause) ref));
				continue;
			} else {
				log.error("Unsupported clause detected: " + ref);
				return false;
			}

			// Naming metrics from undefined aliases makes little sense
			if (entity == null) {
				log.error("Missing entity alias for requested metric, aborting.");
				return false;
			} else {
				final EList<Probe> probes = entity.getProbes();
				probes.add(probe);
			}
		}

		return true;
	}

	private String getDefaultStatType(String entityRef, String metricRef2) {
		List<Entity> entities = new LinkedList<Entity>();
		entities.addAll(result.getResources());
		entities.addAll(result.getServices());
		for (Entity e : entities)
			if (e.getAlias().equals(entityRef))
				for (Probe p : e.getProbes())
					if (p.getMetricName().equals(metricRef2))
						if (p.getStatTypes().isEmpty())
							return "DEFAULT";
						else
							return p.getStatTypes().get(0).getTypeName();
		throw new IllegalStateException("There is no Metric " + entityRef + "." + metricRef2);
	}

	private void initializeEntityMapping(SelectQuery selectQuery) {
		UsingClause uc = selectQuery.getUsingClause();
		ForClause fc = selectQuery.getForClause();

		final DescartesQL metricsQuery = dqlFactory.createDescartesQL();
		final ListQuery listQuery = dqlFactory.createListQuery();
		metricsQuery.setQuery(listQuery);
		listQuery.setUsingClause(uc);
		final ListMetricsQuery listMetricsQuery = dqlFactory.createListMetricsQuery();
		listQuery.setQueryType(listMetricsQuery);
		listMetricsQuery.setMode("LIST METRICS");
		listMetricsQuery.setEntityReferenceClause(fc.getEntityReferenceClause());

		final QueryExecutionEngine qee = new QueryExecutionEngineImpl();
		this.result = qee.execute(metricsQuery).get(0);
	}

	/**
	 * Process a VARYING clause to specify parameters as DoF
	 * 
	 * @param request
	 *            The request to be sent to the query connector
	 * @param varyingClause
	 *            The clause of parameters used as DoF
	 */
	private void processVaryingClause(final EntityMapping request, final VaryingClause varyingClause) {
		for (final DoFReference doFReference : varyingClause.getDofReferences()) {
			final EObject variationClause = doFReference.getDoFVariationClause();
			Properties variationProperties = null;
			final DoF doF = domainFactory.createDoF();
			final EList<DoF> requestDoF = request.getDoF();
			final AliasClause aliasClause = doFReference.getDoFAliasClause();
			
			if(aliasClause != null){
				doF.setAlias(aliasClause.getAlias());				
			}
			
			doF.setIdentifier(doFReference.getDoFIdentifier());

			if (variationClause != null) {
				variationProperties = processVariationClause(variationClause, variationProperties);

				doF.setProperties(variationProperties);
			}

			requestDoF.add(doF);
		}
	}
}
